// Cpp11LambdasVariableCapture.cpp : Defines the entry point for the console application.
//

#include <iostream>
#include <vector>
#include <functional>
#include <algorithm>

void printAll(const std::vector<int> &list) {
	std::for_each(list.begin(), list.end(), [](int x) {
		std::cout << "Value: " << x << std::endl;
	});
}

//
// More examples below!
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//
//

void findMax0(const std::vector<int> &list) {
	int max = INT32_MIN;
	std::for_each(list.begin(), list.end(), [](int x) {
		// Error: 'max' cannot be implicitly captured because no default capture mode has been specified
		// if (x > max) max = x;
	});
	std::cout << "Max0: " << max << std::endl;
}

void printCaptureViaCopy() {
	int value = 123;

	auto func = [=]() {
		std::cout << "printCaptureViaCopy's lambda called: value " << value << std::endl;
	};

	func();

	value = 42;

	func();
}

void printCaptureViaRef() {
	int value = 123;

	auto func = [&]() {
		std::cout << "printCaptureViaRef's lambda called: value " << value << std::endl;
	};

	func();

	value = 42;

	func();
}

void findMax1(const std::vector<int> &list) {
	int max = INT32_MIN;
	std::for_each(list.begin(), list.end(), [=](int x) {
		// Error: 'max': a by copy capture cannot be modified in a non - mutable lambda
		// (error on 'max = x')
		// if (x > max) max = x;
		std::cout << "Iteration: max == " << max << std::endl;
	});
	std::cout << "Max1: " << max << std::endl;
}

void findMax2(const std::vector<int> &list) {
	int max = INT32_MIN;
	std::for_each(list.begin(), list.end(), [&](int x) {
		if (x > max) max = x;
	});
	std::cout << "Max2: " << max << std::endl;
}

std::function<void()> getValueViewer() {
	int value = 123;

	auto func = [=]() {
		std::cout << "getValueViewer's lambda called: value " << value << std::endl;
	};

	func();

	value = 42;

	func();
	return func;
}

std::function<void()> getRefViewer() {
	int value = 123;

	auto func = [&]() {
		std::cout << "getRefViewer's lambda called: value " << value << std::endl;
	};

	func();

	value = 42;

	func();
	return func;
}

std::function<void()> getProcessor() {
	int value = 0;

	auto func = [&]() {
		std::cout << "getProcessor's lambda called: value " << value;
		std::cout << " -> " << ++value << std::endl;
	};

	func();

	std::cout << "getProcessor: value == " << value << std::endl;
	return func;
}

int main()
{
	std::vector<int> numbers;
	numbers.push_back(2);
	numbers.push_back(1);
	numbers.push_back(4);
	numbers.push_back(3);

	printAll(numbers);

	std::cout << "-- Press Enter to continue --";
	std::cin.get();

	std::cout << std::endl << "main: calling findMax0 ..." << std::endl;
	findMax0(numbers);

	std::cout << "-- Press Enter to continue --";
	std::cin.get();

	std::cout << std::endl << "main: calling printCaptureViaCopy ..." << std::endl;
	printCaptureViaCopy();

	std::cout << "-- Press Enter to continue --";
	std::cin.get();

	std::cout << std::endl << "main: calling printCaptureViaRef ..." << std::endl;
	printCaptureViaRef();

	std::cout << "-- Press Enter to continue --";
	std::cin.get();

	std::cout << std::endl << "main: calling findMax1 ..." << std::endl;
	findMax1(numbers);

	std::cout << "-- Press Enter to continue --";
	std::cin.get();

	std::cout << std::endl << "main: calling findMax2 ..." << std::endl;
	findMax2(numbers);

	std::cout << "-- Press Enter to continue --";
	std::cin.get();

	std::cout << std::endl << "main: calling getValueViewer ..." << std::endl;
	auto func = getValueViewer();
	std::cout << "main: calling func ..." << std::endl;
	func();

	std::cout << "-- Press Enter to continue --";
	std::cin.get();

	std::cout << std::endl << "main: calling getRefViewer ..." << std::endl;
	func = getRefViewer();
	std::cout << "main: calling func ..." << std::endl;
	func();

	std::cout << "-- Press Enter to continue --";
	std::cin.get();

	std::cout << std::endl << "main: calling getProcessor ..." << std::endl;
	func = getProcessor();
	std::cout << "main: calling func ..." << std::endl;
	func();

	std::cout << "-- Press Enter to EXIT --";
	std::cin.get();

	return 0;
}

